#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

// Inline ReLU function template
template <typename scalar_t>
__device__ inline scalar_t relu_func(scalar_t x) {
  return x > 0 ? x : 0;
}

template <>
__device__ inline double relu_func(double x) {
  return x > 0 ? x : 0;
}


// This kernel combines a tiled shared memory approach with grid-stride loops
// to cover large matrices while ensuring memory coalescing for both input and weight.
// The kernel fuses a GEMM with bias addition, a subtract, a multiplication, and a ReLU activation.

template <typename scalar_t>
__global__ void tiled_grid_stride_matmul_relu_kernel(
    const scalar_t* __restrict__ input,    // [batch_size x in_features]
    const scalar_t* __restrict__ weight,   // [out_features x in_features]
    const scalar_t* __restrict__ bias,     // [out_features]
    scalar_t* __restrict__ output,         // [batch_size x out_features]
    const int batch_size,
    const int in_features,
    const int out_features,
    const float subtract_val,
    const float multiply_val) {

  // Tile dimension (adjustable) 
  constexpr int TILE_DIM = 16;

  // Allocate shared memory for input and weight tiles
  __shared__ scalar_t shInput[TILE_DIM][TILE_DIM];
  __shared__ scalar_t shWeight[TILE_DIM][TILE_DIM];

  // Use grid-stride loops to cover the full output matrix
  for (int row_tile = blockIdx.y * TILE_DIM;
       row_tile < batch_size;
       row_tile += gridDim.y * TILE_DIM) {
    for (int col_tile = blockIdx.x * TILE_DIM;
         col_tile < out_features;
         col_tile += gridDim.x * TILE_DIM) {

      // Compute the output coordinate this thread is responsible for
      int row = row_tile + threadIdx.y;
      int col = col_tile + threadIdx.x;
      scalar_t acc = 0;

      // Number of tiles along the in_features (K) dimension
      int numTiles = (in_features + TILE_DIM - 1) / TILE_DIM;

      // Loop over tiles in the K dimension
      for (int t = 0; t < numTiles; t++) {
        int k_index = t * TILE_DIM + threadIdx.x;
        // Load input tile element (each thread loads one element)
        if (row < batch_size && k_index < in_features) {
          shInput[threadIdx.y][threadIdx.x] = input[row * in_features + k_index];
        } else {
          shInput[threadIdx.y][threadIdx.x] = 0;
        }

        int k_index_weight = t * TILE_DIM + threadIdx.y;
        // Load weight tile element using __ldg for read-only cache efficiency
        if (col < out_features && k_index_weight < in_features) {
          shWeight[threadIdx.y][threadIdx.x] = __ldg(&weight[col * in_features + k_index_weight]);
        } else {
          shWeight[threadIdx.y][threadIdx.x] = 0;
        }

        __syncthreads();

        // Compute partial dot-product for the current tile
        #pragma unroll
        for (int k = 0; k < TILE_DIM; k++) {
          acc += shInput[threadIdx.y][k] * shWeight[k][threadIdx.x];
        }

        __syncthreads();
      } // End of K-dimension tiling loop

      // After accumulation, if within bounds, apply bias, subtract, multiply, and ReLU
      if (row < batch_size && col < out_features) {
        scalar_t value = acc + bias[col];
        value = (value - subtract_val) * multiply_val;
        // Using the inline ReLU
        output[row * out_features + col] = relu_func(value);
      }
    } // end for col_tile
  } // end for row_tile
}

// Host-side forward function exposed to PyTorch

torch::Tensor forward(
    torch::Tensor input,
    torch::Tensor weight,
    torch::Tensor bias,
    float subtract_val,
    float multiply_val) {

  auto batch_size = input.size(0);
  auto in_features = input.size(1);
  auto out_features = weight.size(0);

  auto output = torch::empty({batch_size, out_features}, input.options());

  constexpr int TILE_DIM = 16;

  // Define 2D block and grid dimensions. Grid-stride loops ensure full coverage even if grid size is smaller than problem dimensions.
  dim3 threads(TILE_DIM, TILE_DIM);
  dim3 blocks((out_features + TILE_DIM - 1) / TILE_DIM, (batch_size + TILE_DIM - 1) / TILE_DIM);

  AT_DISPATCH_FLOATING_TYPES(input.scalar_type(), "tiled_grid_stride_matmul_relu_kernel", ([&] {
    tiled_grid_stride_matmul_relu_kernel<scalar_t><<<blocks, threads>>>(
        input.data_ptr<scalar_t>(),
        weight.data_ptr<scalar_t>(),
        bias.data_ptr<scalar_t>(),
        output.data_ptr<scalar_t>(),
        batch_size,
        in_features,
        out_features,
        subtract_val,
        multiply_val);
  }));

  return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
  m.def("forward", &forward, "Tiled and grid-stride fused matmul with bias, subtract, multiply and ReLU (CUDA)");
}
